@@ -10,7 +10,8 @@ import functools
 import sys
 from IPython.display import display
 import random
-
+import io
+# import progressbar
 from object_detection.utils import dataset_util
 from object_detection.utils import config_util
 from object_detection.utils import label_map_util
@@ -29,6 +30,10 @@ from object_detection import model_lib_v2
 from tensorflow.python.summary.summary_iterator import summary_iterator
 from google.protobuf import text_format
 from object_detection.protos import optimizer_pb2
+from object_detection.core import data_parser
+from object_detection.core import standard_fields as fields
+from object_detection.metrics.tf_example_parser import BoundingBoxParser, StringParser, Int64Parser, FloatParser
+
 
 def create_record_files(dict_annotations, train_list, train_list_id, eval_list, 
                         eval_list_id, label_path, record_dir, tfExample_generator, annotation_type):
@@ -454,105 +459,7 @@ def train(ckpt_dir='', config_dir='', train_steps=None, use_tpu=False, checkpoin
           use_tpu=use_tpu,
           checkpoint_every_n=checkpoint_every_n,
           record_summaries=record_summaries)
-# def train(master='', save_summaries_secs=30, task=0, num_clones=1, clone_on_cpu=False, worker_replicas=1, ps_tasks=0, 
-#                     ckpt_dir='', conf_dir='', train_config_path='', input_config_path='', model_config_path=''):   
-   
 
-#     pipeline_config_path = os.path.join(conf_dir,"pipeline.config")
-#     configs = config_util.get_configs_from_pipeline_file(pipeline_config_path)
-
-#     tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.INFO)
-#     assert ckpt_dir, '`ckpt_dir` is missing.'
-#     if task == 0: tf.compat.v1.gfile.MakeDirs(ckpt_dir)
-#     if pipeline_config_path:
-#         configs = config_util.get_configs_from_pipeline_file(pipeline_config_path)
-#         if task == 0:
-#             tf.compat.v1.gfile.Copy(pipeline_config_path,
-#                           os.path.join(ckpt_dir, 'pipeline.config'),
-#                           overwrite=True)
-#     else:
-#         configs = config_util.get_configs_from_multiple_files(
-#                         model_config_path=model_config_path,
-#                         train_config_path=train_config_path,
-#                         train_input_config_path=input_config_path)
-#         if task == 0:
-#             for name, config in [('model.config', model_config_path),
-#                                 ('train.config', train_config_path),
-#                                 ('input.config', input_config_path)]:
-#                 tf.compat.v1.gfile.Copy(config, os.path.join(ckpt_dir, name),
-#                             overwrite=True)
-
-#     model_config = configs['model']
-#     train_config = configs['train_config']
-#     input_config = configs['train_input_config']
-
-#     model_fn = functools.partial(
-#         model_builder.build,
-#         model_config=model_config,
-#         is_training=True)
-
-#     def get_next(config):
-#         return dataset_builder.make_initializable_iterator(
-#             dataset_builder.build(config)).get_next()
-
-#     create_input_dict_fn = functools.partial(get_next, input_config)
-
-#     env = json.loads(os.environ.get('TF_CONFIG', '{}'))
-#     cluster_data = env.get('cluster', None)
-#     cluster = tf.train.ClusterSpec(cluster_data) if cluster_data else None
-#     task_data = env.get('task', None) or {'type': 'master', 'index': 0}
-#     task_info = type('TaskSpec', (object,), task_data)
-
-#     # Parameters for a single worker.
-#     ps_tasks = 0
-#     worker_replicas = 1
-#     worker_job_name = 'lonely_worker'
-#     task = 0
-#     is_chief = True
-#     master = ''
-
-#     if cluster_data and 'worker' in cluster_data:
-#     # Number of total worker replicas include "worker"s and the "master".
-#         worker_replicas = len(cluster_data['worker']) + 1
-#     if cluster_data and 'ps' in cluster_data:
-#         ps_tasks = len(cluster_data['ps'])
-
-#     if worker_replicas > 1 and ps_tasks < 1:
-#         raise ValueError('At least 1 ps task is needed for distributed training.')
-
-#     if worker_replicas >= 1 and ps_tasks > 0:
-#     # Set up distributed training.
-#         server = tf.train.Server(tf.train.ClusterSpec(cluster), protocol='grpc',
-#                                 job_name=task_info.type,
-#                                 task_index=task_info.index)
-#         if task_info.type == 'ps':
-#             server.join()
-#             return
-
-#         worker_job_name = '%s/task:%d' % (task_info.type, task_info.index)
-#         task = task_info.index
-#         is_chief = (task_info.type == 'master')
-#         master = server.target
-
-#     graph_rewriter_fn = None
-#     if 'graph_rewriter_config' in configs:
-#         graph_rewriter_fn = graph_rewriter_builder.build(
-#             configs['graph_rewriter_config'], is_training=True)
-
-#     trainer.train(
-#         create_input_dict_fn,
-#         model_fn,
-#         train_config,
-#         master,
-#         task,
-#         num_clones,
-#         worker_replicas,
-#         clone_on_cpu,
-#         ps_tasks,
-#         worker_job_name,
-#         is_chief,
-#         ckpt_dir,
-#         graph_hook_fn=graph_rewriter_fn)
 
 def evaluate(metrics_dir='', config='', ckpt_dir=''):
     '''
@@ -741,4 +648,200 @@ def infer(path, exported_model_dir, label_map_path, results_dir, disp=True, num_
         print(f"{counter} weren't infered")
 
 def run():
-    pass
\ No newline at end of file
+    pass
+
+
+FLAGS = tf.compat.v1.flags.FLAGS
+
+class CustomParser(data_parser.DataToNumpyParser):
+    """Tensorflow Example proto parser."""
+
+    def __init__(self):
+        self.items_to_handlers = {
+            fields.InputDataFields.image: StringParser(
+                        fields.TfExampleFields.image_encoded),
+            fields.InputDataFields.groundtruth_boxes: BoundingBoxParser(
+                        fields.TfExampleFields.object_bbox_xmin,
+                        fields.TfExampleFields.object_bbox_ymin,
+                        fields.TfExampleFields.object_bbox_xmax,
+                        fields.TfExampleFields.object_bbox_ymax),
+            fields.InputDataFields.groundtruth_classes: Int64Parser(
+                        fields.TfExampleFields.object_class_label)
+        }
+        # Optional
+        self.filename = StringParser(fields.TfExampleFields.filename)
+
+    def parse(self, tf_example):
+        """
+        Parses tensorflow example and returns a tensor dictionary.
+        Args:
+            tf_example: a tf.Example object.
+        Returns:
+            A dictionary of the following numpy arrays:
+            image               - string containing input image.
+            filename            - string containing input filename (optional, None if not specified)
+            groundtruth_boxes   - a numpy array containing groundtruth boxes.
+            groundtruth_classes - a numpy array containing groundtruth classes.
+        """
+        results_dict = {}
+        parsed = True
+        for key, parser in self.items_to_handlers.items():
+            results_dict[key] = parser.parse(tf_example)
+            parsed &= (results_dict[key] is not None)
+
+        # TODO: need to test
+        filename = self.filename.parse(tf_example)
+        results_dict['filename'] = filename # could be None
+
+        return results_dict if parsed else None
+
+def compute_iou(groundtruth_box, detection_box):
+    g_ymin, g_xmin, g_ymax, g_xmax = tuple(groundtruth_box.tolist())
+    d_ymin, d_xmin, d_ymax, d_xmax = tuple(detection_box.tolist())
+    
+    xa = max(g_xmin, d_xmin)
+    ya = max(g_ymin, d_ymin)
+    xb = min(g_xmax, d_xmax)
+    yb = min(g_ymax, d_ymax)
+
+    intersection = max(0, xb - xa + 1) * max(0, yb - ya + 1)
+
+    boxAArea = (g_xmax - g_xmin + 1) * (g_ymax - g_ymin + 1)
+    boxBArea = (d_xmax - d_xmin + 1) * (d_ymax - d_ymin + 1)
+
+    return intersection / float(boxAArea + boxBArea - intersection)
+
+
+def get_confusion_matrix(input_tfrecord_path, model, labelmap, iou_thresh=0.5, confidence_thresh=0.5, draw_option=False, draw_save_path=""):
+    """
+    Creates input dataset from tfrecord, runs detection model, compares detection results with ground truth
+    Args:
+        input_tfrecord_path: path of input tfrecord file
+        model: path of detection model .pb file
+        categories: ordered array of class IDs
+        draw_option: whether or not to visualize and save detections and ground truth boxes
+        draw_save_path: where to save visualizations if draw_option is true
+    """
+    IOU_THRESHOLD = iou_thresh
+    CONFIDENCE_THRESHOLD = confidence_thresh
+    data_parser = CustomParser()
+    categories = [lab for lab in labelmap.keys()]
+    labels = [lab for lab in labelmap.values()]
+    confusion_matrix = np.zeros(shape=(len(categories) + 1, len(categories) + 1))
+    filename_matrix = [[[] for i in range(len(categories)+1)] for j in range(len(categories)+1)]
+    # Create dataset from records
+    input_dataset = tf.data.TFRecordDataset(input_tfrecord_path)
+    # model = tf.saved_model.load(model_path)
+    # with progressbar.ProgressBar(max_value=progressbar.UnknownLength) as bar:
+    results = []
+    
+    total_objects= np.zeros(len(categories))
+    for image_index, record in enumerate(input_dataset):
+        if image_index < 3:
+            example = tf.train.Example()
+            example.ParseFromString(record.numpy())
+            decoded_dict = data_parser.parse(example)
+            if decoded_dict:
+                image               = decoded_dict[fields.InputDataFields.image]
+                input_tensor        = np.expand_dims(Image.open(io.BytesIO(image)), axis=0)
+                groundtruth_boxes   = decoded_dict[fields.InputDataFields.groundtruth_boxes]
+                groundtruth_classes = decoded_dict[fields.InputDataFields.groundtruth_classes].astype('uint8')
+                detections          = model(input_tensor) # Run model inference
+                detection_scores    = detections['detection_scores'][0].numpy()
+                detection_boxes     = detections['detection_boxes'][0].numpy()[detection_scores >= CONFIDENCE_THRESHOLD]
+                detection_classes   = detections['detection_classes'][0].numpy()[detection_scores >= CONFIDENCE_THRESHOLD].astype('uint8')
+                filename            = decoded_dict[fields.InputDataFields.filename]
+                filename            = filename.decode('UTF-8') if filename is not None else f'image-{image_index}.png'
+                print(filename)
+                matches = []
+                image_index += 1
+                gt_boxes = []
+                pred_boxes = []
+                for i, groundtruth_box in enumerate(groundtruth_boxes):
+                    for j, detection_box in enumerate(detection_boxes):
+                        iou = compute_iou(groundtruth_box, detection_box)
+                        if iou > IOU_THRESHOLD:
+                            matches.append([i, j, iou])
+                            gt_boxes.append(groundtruth_boxes)
+                            pred_boxes.append(detection_boxes)
+                matches = np.array(matches)
+                if matches.shape[0] > 0:
+                    # Sort list of matches by descending IOU so we can remove duplicate detections
+                    # while keeping the highest IOU entry.
+                    matches = matches[matches[:, 2].argsort()[::-1][:len(matches)]]
+                    
+                    # Remove duplicate detections from the list.
+                    matches = matches[np.unique(matches[:,1], return_index=True)[1]]
+                    
+                    # Sort the list again by descending IOU. Removing duplicates doesn't preserve
+                    # our previous sort.
+                    matches = matches[matches[:, 2].argsort()[::-1][:len(matches)]]
+                    
+                    # Remove duplicate ground truths from the list.
+                    matches = matches[np.unique(matches[:,0], return_index=True)[1]]
+                print(matches)
+                print(groundtruth_classes)
+                print(detection_classes)
+                nb_objects = np.zeros(len(categories))
+                nb_preds = np.zeros(len(categories))
+                FN = np.zeros(len(categories))
+                FP = np.zeros(len(categories))
+                TP = np.zeros(len(categories))
+                for i in range(len(groundtruth_boxes)):
+                    if matches.shape[0] > 0 and matches[matches[:,0] == i].shape[0] == 1:
+                        confusion_matrix[groundtruth_classes[i] - 1][int(detection_classes[int(matches[matches[:,0] == i, 1][0])] - 1)] += 1
+                        if filename not in filename_matrix[groundtruth_classes[i] - 1][int(detection_classes[int(matches[matches[:,0] == i, 1][0])] - 1)]:
+                            filename_matrix[groundtruth_classes[i] - 1][int(detection_classes[int(matches[matches[:,0] == i, 1][0])] - 1)].append(filename)
+                        if int(detection_classes[int(matches[matches[:,0] == i, 1][0])] - 1) == groundtruth_classes[i] - 1:
+                            TP[groundtruth_classes[i]-1] += 1
+                        else:
+                            FP[int(detection_classes[int(matches[matches[:,0] == i, 1][0])] - 1)] += 1
+                    else:
+                        confusion_matrix[groundtruth_classes[i] - 1][confusion_matrix.shape[1] - 1] += 1
+                        if filename not in filename_matrix[groundtruth_classes[i] - 1][confusion_matrix.shape[1] - 1]:
+                            filename_matrix[groundtruth_classes[i] - 1][confusion_matrix.shape[1] - 1].append(filename)
+                        FN[groundtruth_classes[i]-1] += 1
+                    total_objects[groundtruth_classes[i]-1] += 1
+                    nb_objects[groundtruth_classes[i]-1] += 1
+
+                for i in range(len(detection_boxes)):
+                    if matches.shape[0] > 0 and matches[matches[:,1] == i].shape[0] == 0:
+                        confusion_matrix[confusion_matrix.shape[0] - 1][int(detection_classes[i] - 1)] += 1
+                        if filename not in filename_matrix[confusion_matrix.shape[0] - 1][int(detection_classes[i] - 1)]:
+                            filename_matrix[confusion_matrix.shape[0] - 1][int(detection_classes[i] - 1)].append(filename)
+                        FP[detection_classes[i]-1] += 1
+                    nb_preds[detection_classes[i]-1] += 1
+
+                image_dict = {
+                    'filename': filename.split('/')[-1],
+                    'nb_objects': nb_objects.tolist(),
+                    'nb_preds': nb_preds.tolist(),
+                    'TP': TP.tolist(),
+                    'FP': FP.tolist(),
+                    'FN': FN.tolist(),
+                    'bbox': detection_boxes.tolist(),
+                    'classes': detection_classes.tolist()
+                }
+                results.append(image_dict)
+            # if draw_option:
+            #     draw(filename, draw_save_path, input_tensor[0], categories,
+            #         groundtruth_boxes, groundtruth_classes, detection_boxes, detection_classes, detection_scores)
+            else:
+                print(f'Skipping image {image_index}')
+
+            # bar.update(image_index)
+
+    print(f'Processed {image_index + 1} images')
+    evaluation = {
+        'categories': categories,
+        'labels': labels,
+        'total_objects': total_objects.tolist(),
+        'per_image_eval': results,
+        'filename_matrix': filename_matrix,
+        'metrics': {
+            'iou': IOU_THRESHOLD,
+            'confidence': CONFIDENCE_THRESHOLD,
+        }
+    }
+    return confusion_matrix, evaluation
+
